import numpy as np


def calculate_confusion_matrix_from_arrays(prediction, ground_truth, nr_labels):
    """
    https://github.com/ternaus/robot-surgery-segmentation/blob/master/validation.py
    """

    replace_indices = np.vstack((
        ground_truth.flatten(),
        prediction.flatten())
    ).T

    confusion_matrix, _ = np.histogramdd(
        replace_indices,
        bins=(nr_labels, nr_labels),
        range=[(0, nr_labels), (0, nr_labels)]
    )
    confusion_matrix = confusion_matrix.astype(np.uint64)
    return confusion_matrix


def calculate_dice(confusion_matrix):
    """
    https://github.com/ternaus/robot-surgery-segmentation/blob/master/validation.py
    """
    confusion_matrix = confusion_matrix.astype(float)
    dices = []
    for index in range(confusion_matrix.shape[0]):
        true_positives = confusion_matrix[index, index]
        false_positives = confusion_matrix[:, index].sum() - true_positives
        false_negatives = confusion_matrix[index, :].sum() - true_positives
        denom = 2 * true_positives + false_positives + false_negatives
        if denom == 0:
            dice = 0
        else:
            dice = 2 * float(true_positives) / denom
        dices.append(dice)
    return dices


def calculate_iou(confusion_matrix):
    """
    https://github.com/ternaus/robot-surgery-segmentation/blob/master/validation.py
    """
    confusion_matrix = confusion_matrix.astype(float)
    ious = []
    for index in range(confusion_matrix.shape[0]):
        true_positives = confusion_matrix[index, index]
        false_positives = confusion_matrix[:, index].sum() - true_positives
        false_negatives = confusion_matrix[index, :].sum() - true_positives
        denom = true_positives + false_positives + false_negatives
        if denom == 0:
            iou = 0
        else:
            iou = float(true_positives) / denom
        ious.append(iou)
    return ious


def calculate_volumetric_similarity(confusion_matrix):
    """
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4533825/

    """
    confusion_matrix = confusion_matrix.astype(float)
    scores = []
    for index in range(confusion_matrix.shape[0]):
        true_positives = confusion_matrix[index, index]
        false_positives = confusion_matrix[:, index].sum() - true_positives
        false_negatives = confusion_matrix[index, :].sum() - true_positives
        denom = 2 * true_positives + false_positives + false_negatives
        if denom == 0:
            vd = 0
        else:
            vd = 1 - abs(false_negatives-false_positives) / denom

        scores.append(vd)

    return scores


import torch
from typing import Callable
import numpy as np
import time
from termcolor import colored
import subprocess
import socket
from torch import optim
import os
import operator
from deeppipeline.kvs import GlobalKVS
from tensorboardX import SummaryWriter
from torch.optim import lr_scheduler


def git_info():
    """
    Gathers info about the version of the codebase from git.

    Returns
    -------
    out : tuple of str
        branch and commit id
    """
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        git_revision = out.strip().decode('ascii')

        out = _minimal_ext_cmd(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])
        git_branch = out.strip().decode('ascii')
    except OSError:
        return None

    return git_branch, git_revision


def init_session(args):
    """
    Basic function that initializes each training loop.
    Sets the seed based on the parsed args, creates the snapshots dir and initializes global KVS.

    Parameters
    ----------
    args : Namespace
        Arguments from argparse.

    Returns
    -------
    out : tuple
        Args, snapshot name and global KVS.
    """
    # Initializing the seeds
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed(args.seed)
    np.random.seed(args.seed)
    # Creating the snapshot
    snapshot_name = time.strftime(f'{socket.gethostname()}_%Y_%m_%d_%H_%M')
    os.makedirs(os.path.join(args.workdir, 'snapshots',  snapshot_name), exist_ok=True)

    kvs = GlobalKVS(os.path.join(args.workdir, 'snapshots', snapshot_name, 'session.pkl'))

    res = git_info()
    if res is not None:
        kvs.update('git branch name', res[0])
        kvs.update('git commit id', res[1])
    else:
        kvs.update('git branch name', None)
        kvs.update('git commit id', None)

    kvs.update('pytorch_version', torch.__version__)

    if torch.cuda.is_available():
        kvs.update('cuda', torch.version.cuda)
        kvs.update('gpus', torch.cuda.device_count())
    else:
        kvs.update('cuda', None)
        kvs.update('gpus', None)

    kvs.update('snapshot_name', snapshot_name)
    kvs.update('args', args)

    return args, snapshot_name, kvs


def init_optimizer(net):
    """
    Initializes the optimizer for a given model.
    Currently supported optimizers are Adam and SGD with default parameters.
    Learning rate (LR) and weight decay (WD) must be specified in the arguments as `lr` and `wd`, respectively.
    LR and WD are retrieved automatically from global KVS.
    Parameters
    ----------
    net : torch.Module

    Returns
    -------
    out : torch.optim.Optimizer
        Initialized optimizer.

    """
    kvs = GlobalKVS()
    if kvs['args'].optimizer == 'adam':
        return optim.Adam(net.parameters(), lr=kvs['args'].lr, weight_decay=kvs['args'].wd)
    elif kvs['args'].optimizer == 'sgd':
        return optim.SGD(net.parameters(), lr=kvs['args'].lr, weight_decay=kvs['args'].wd, momentum=0.9)
    else:
        raise NotImplementedError


def log_metrics(writer, train_loss, val_loss, val_results, val_results_callback=None):
    """
    Basic function to log the results from the validation stage.
    takes Tensorboard writer, train loss, validation loss, the artifacts produced during the validation phase,
    and also additional callback that can process these data, e.g. compute the metrics and
    visualize them in Tensorboard. By default, train and validation losses are visualized outside of the callback.
    If any metric is computed in the callback, it is useful to log it into a dictionary `to_log`.



    Parameters
    ----------
    writer : SummaryWriter
        Tensorboard summary writer
    train_loss : float
        Training loss
    val_loss : float
        Validation loss
    val_results : object
        Artifacts produced during teh validation
    val_results_callback : Callable or None
        A callback function that can process the artifacts and e.g. display those in Tensorboard.

    Returns
    -------
    out : None

    """
    kvs = GlobalKVS()

    print(colored('==> ', 'green') + 'Metrics:')
    print(colored('====> ', 'green') + 'Train loss:', train_loss)
    print(colored('====> ', 'green') + 'Val loss:', val_loss)

    to_log = {'train_loss': train_loss, 'val_loss': val_loss}
    val_metrics = {'epoch': kvs['cur_epoch']}
    val_metrics.update(to_log)
    writer.add_scalars(f"Losses_{kvs['args'].annotations}", to_log, kvs['cur_epoch'])
    if val_results_callback is not None:
        val_results_callback(writer, val_metrics, to_log, val_results)

    kvs.update(f'losses_fold_[{kvs["cur_fold"]}]', to_log)
    kvs.update(f'val_metrics_fold_[{kvs["cur_fold"]}]', val_metrics)


def init_ms_scheduler(optimizer):
    """
    Initializes a simple multi-step learning rate scheduler.
    The scheduling is done according to the scheduling parameters specified in the arguments.
    The parameter responsible for this is `lr_drop`.

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Optimizer for which the scheduler need to be created
    Returns
    -------
    out : lr_scheduler.Scheduler
        Created Scheduler

    """
    kvs = GlobalKVS()
    return lr_scheduler.MultiStepLR(optimizer, kvs['args'].lr_drop)


def save_checkpoint(net, optimizer, val_metric_name, comparator='lt'):
    """
    Flexible function that saves the model and the optimizer states using a metric and a comparator.

    Parameters
    ----------
    net : torch.nn.Module
        Model
    optimizer : torch.optim.Optimizer
        Optimizer
    val_metric_name : str
        Name of the metric that needs to be used for snapshot comparison.
        This name needs match the once that were created in the callback function passed to
        `log_metrics`.
    comparator : str
        How to compare the previous and the current metric values - `lt` is less than, and `gt` is greater than.

    Returns
    -------
    out : None

    """
    if isinstance(net, torch.nn.DataParallel):
        net = net.module

    kvs = GlobalKVS()
    fold_id = kvs['cur_fold']
    epoch = kvs['cur_epoch']
    val_metric = kvs[f'val_metrics_fold_[{fold_id}]'][-1][0][val_metric_name]
    comparator = getattr(operator, comparator)
    cur_snapshot_name = os.path.join(os.path.join(kvs['args'].workdir, 'snapshots', kvs['snapshot_name'],
                                     f'fold_{fold_id}_epoch_{epoch}.pth'))

    if kvs['prev_model'] is None:
        print(colored('====> ', 'red') + 'Snapshot was saved to', cur_snapshot_name)
        torch.save({'model': net.state_dict(), 'optimizer': optimizer.state_dict()}, cur_snapshot_name)
        kvs.update('prev_model', cur_snapshot_name)
        kvs.update('best_val_metric', val_metric)

    else:
        if comparator(val_metric, kvs['best_val_metric']):
            print(colored('====> ', 'red') + 'Snapshot was saved to', cur_snapshot_name)
            os.remove(kvs['prev_model'])
            torch.save({'model': net.state_dict(), 'optimizer': optimizer.state_dict()}, cur_snapshot_name)
            kvs.update('prev_model', cur_snapshot_name)
            kvs.update('best_val_metric', val_metric)


def train_fold(pass_epoch, net, train_loader, optimizer, criterion, val_loader, scheduler, log_metrics_cb=None):
    """
    A common implementation of training one fold of a neural network. Presumably, it should be called
    within cross-validation loop.

    Parameters
    ----------
    pass_epoch : Callable
        Function that trains or validates one epoch
    net : torch.nn.Module
        Model to train
    train_loader : torch.utils.data.DataLoader
        Training data loader
    optimizer : torch.optim.Optimizer
        Optimizer
    criterion : torch.nn.Module
        Loss function
    val_loader : torch.utils.data.DataLoader
        Validation data loader
    scheduler : lr_scheduler.Scheduler
        Learning rate scheduler
    log_metrics_cb : Callable or None
        Callback that processes the artifacts from validation stage.

    Returns
    -------

    """
    kvs = GlobalKVS()
    fold_id = kvs['cur_fold']
    writer = SummaryWriter(os.path.join(kvs['args'].workdir, 'snapshots', kvs['snapshot_name'],
                                        'logs', 'fold_{}'.format(fold_id), kvs['snapshot_name']))

    for epoch in range(kvs['args'].n_epochs):
        print(colored('==> ', 'green') + f'Training epoch [{epoch}] with LR {scheduler.get_lr()}')
        kvs.update('cur_epoch', epoch)
        train_loss, _ = pass_epoch(net, train_loader, optimizer, criterion)
        val_loss, val_results = pass_epoch(net, val_loader, None, criterion)
        log_metrics(writer, train_loss, val_loss, val_results, log_metrics_cb)
        save_checkpoint(net, optimizer, 'val_loss', 'lt')
        scheduler.step()


def train_n_folds(init_args, init_metadata, init_augs,
                  init_data_processing,
                  init_folds,
                  init_loaders,
                  init_model, init_loss,
                  init_scheduler,
                  pass_epoch, log_metrics_cb,
                  img_group_id_colname=None, img_class_colname=None):
    """
    Full implementation of the n-fold training loop. Trains each fold in cross-validation.

    Parameters
    ----------
    init_args : Callable
        Function that initializes the arguments. This should store the arguments
        in global KVS as `args`.
    init_metadata : Callable
        Function that initializes metadata as a pandas dataframe. The function must store this object
        in global KVS under `metadata` tag.
    init_augs : Callable
        Must initialize the training data augmentations. Should store the callable transforms
        as `train_trf` in global KVS. Train transforms should be returned **without** mean-std normaliztion.
    init_data_processing : Callable
        Initializes the data processing. Should compute the mean and std or load those from file.
        Typically, this leverages the existing function from deep-pipeline see `deeppipeline.common.core.normaliztion`.
        After the mean and std are initialized they can be appended to `train_trf` and `train_trf` must be updated in
        global kvs. Thi function should also initialize the validation transforms.
    init_folds : Callable
        Initializes the n-fold cross validation depending on whether class and group for each image are specified.
    init_loaders : Callable
        Initializes the data loaders given the train and validation data splits stored as pandas data frames.
        This function is executed for every fold.
    init_model : Callable
        Initializes the model for every fold.
    init_loss : Callable
        Initializes the loss
    init_scheduler : Callable
        Initializes the scheduler
    pass_epoch : Callable
        Trains / validaties one epoch.
    log_metrics_cb : Callable or None
        Callback to process the artifacts of the validation function.
    img_group_id_colname : str or None
        Group id for each image. If not None, the folds are generated so that the images
        with the same group id cannot be in both train and validation simultaneously.
    img_class_colname : str or None
        Class for each image (if any specified). Used to stratify the train and the validation splits
        for every cross-validation fold.

    Returns
    -------
    out : None

    """

    args = init_args()
    kvs = init_session(args)[-1]
    init_metadata()
    assert 'metadata' in kvs

    if init_augs is None:
        raise NotImplementedError('Train augmentations are not defined !!!')
    else:
        init_augs()
    assert 'train_trf' in kvs
    init_data_processing()
    init_folds(img_group_id_colname=img_group_id_colname, img_class_colname=img_class_colname)

    for fold_id, x_train, x_val in kvs['cv_split']:
        kvs.update('cur_fold', fold_id)
        kvs.update('prev_model', None)

        net = init_model()
        optimizer = init_optimizer(net)
        criterion = init_loss()
        scheduler = init_scheduler(optimizer)
        train_loader, val_loader = init_loaders(x_train, x_val)

        train_fold(pass_epoch=pass_epoch, net=net, train_loader=train_loader,
                   optimizer=optimizer, criterion=criterion,
                   val_loader=val_loader, scheduler=scheduler, log_metrics_cb=log_metrics_cb)



from torch import nn
import torch
import torch.nn.functional as F


class BCEWithLogitsLoss2d(nn.Module):
    """Computationally stable version of 2D BCE loss

    """

    def __init__(self):
        super(BCEWithLogitsLoss2d, self).__init__()

        self.bce_loss = nn.BCEWithLogitsLoss(None, reduction='mean')

    def forward(self, logits, targets):
        logits_flat = logits.view(-1)
        targets_flat = targets.view(-1)
        return self.bce_loss(logits_flat, targets_flat)


class SoftJaccardLoss(nn.Module):
    """SoftJaccard loss

    """

    def __init__(self, use_log=False):
        super(SoftJaccardLoss, self).__init__()
        self.use_log = use_log

    def forward(self, logits, labels):
        num = labels.size(0)
        m1 = torch.sigmoid(logits.view(num, -1))
        m2 = labels.view(num, -1)
        intersection = (m1 * m2).sum(1)
        score = (intersection + 1e-15) / (m1.sum(1) + m2.sum(1) - intersection + 1e-15)
        jaccard = score.sum(0) / num

        if not self.use_log:
            score = 1 - jaccard
        else:
            score = -torch.log(jaccard)
        return score


class CombinedLoss(nn.Module):
    """Combination loss.

    Used to combine several existing losses, e.g. Dice and BCE

    """

    def __init__(self, losses, weights=None):
        super(CombinedLoss, self).__init__()
        self.losses = losses
        if weights is None:
            weights = [1 / len(losses)] * len(losses)

        self.weights = weights

    def forward(self, inputs, targets):
        loss = 0
        for l, w in zip(self.losses, self.weights):
            loss += l(inputs, targets) * w
        return loss


class FocalLoss(nn.Module):
    """
    Focal loss. Based on the implementation by BloodAxe:
    https://github.com/BloodAxe/pytorch-toolbelt/blob/develop/pytorch_toolbelt/losses/functional.py#L8

    """
    def __init__(self, gamma=2.0, alpha=0.25, reduction='mean'):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = reduction

    def forward(self, logits, targets):
        num = logits.size(0)
        logits = logits.view(num, -1)
        targets = targets.view(num, -1)
        logpt = -F.binary_cross_entropy_with_logits(logits, targets, reduction='none')
        pt = torch.exp(logpt)

        # compute the loss
        loss = -((1 - pt).pow(self.gamma)) * logpt

        if self.alpha is not None:
            loss = loss * (self.alpha * targets + (1 - self.alpha) * (1 - targets))

        if self.reduction == 'mean':
            loss = loss.mean()
        if self.reduction == 'sum':
            loss = loss.sum()
        if self.reduction == 'batchwise_mean':
            loss = loss.sum(0)

        return loss



from torch import nn
from deeppipeline.kvs import GlobalKVS
from ._unet import UNet


def init_model(ignore_data_parallel=False):
    kvs = GlobalKVS()
    if kvs['args'].model == 'unet':
        net = UNet(bw=kvs['args'].bw, depth=kvs['args'].depth,
                   center_depth=kvs['args'].cdepth,
                   n_inputs=kvs['args'].n_inputs,
                   n_classes=kvs['args'].n_classes - 1,
                   activation='relu')
    else:
        raise NotImplementedError

    if not ignore_data_parallel:
        if kvs['gpus'] > 1:
            net = nn.DataParallel(net).to('cuda')

    net = net.to('cuda')

    return net



from tqdm import tqdm
import numpy as np
import os
from termcolor import colored
from torch.utils.data import DataLoader, Dataset
import torch


def init_mean_std(snapshots_dir: str, dataset: Dataset, batch_size: int, n_threads: int, n_classes: int):
    """
    Calculates mean and std for the dataset. If the masks are available, calculates also the class weights.

    Parameters
    ----------
    snapshots_dir
    dataset
    batch_size
    n_threads
    n_classes

    Returns
    -------

    """
    if os.path.isfile(os.path.join(snapshots_dir, 'mean_std.npy')):
        return np.load(os.path.join(snapshots_dir, 'mean_std.npy'), allow_pickle=True)
    else:
        tmp_loader = DataLoader(dataset, batch_size=batch_size, num_workers=n_threads)
        mean_vector = None
        std_vector = None
        num_pixels = None
        class_weights = None
        print(colored('==> ', 'green') + 'Calculating mean and std')

        if len(tmp_loader) == 0:
            raise ValueError('The data loader has no elements!')

        for batch in tqdm(tmp_loader, total=len(tmp_loader)):
            imgs = batch['img']

            if mean_vector is None:
                mean_vector = np.zeros(imgs.size(1))
                std_vector = np.zeros(imgs.size(1))

            for j in range(mean_vector.shape[0]):
                mean_vector[j] += imgs[:, j, :, :].mean()
                std_vector[j] += imgs[:, j, :, :].std()

            if 'mask' in batch:
                masks = batch['mask']
                if class_weights is None:
                    class_weights = np.zeros(n_classes)
                    num_pixels = 0

                for j in range(class_weights.shape[0]):
                    class_weights[j] += np.sum(masks.numpy() == j)
                num_pixels += np.prod(masks.size())

        mean_vector /= len(tmp_loader)
        std_vector /= len(tmp_loader)
        if class_weights is not None:
            class_weights /= num_pixels
            class_weights = 1 / class_weights
            class_weights /= class_weights.max()
            to_save = [mean_vector.astype(np.float32), std_vector.astype(np.float32), class_weights.astype(np.float32)]
        else:
            to_save = [mean_vector.astype(np.float32), std_vector.astype(np.float32)]

        np.save(os.path.join(snapshots_dir, 'mean_std.npy'), to_save)

    return to_save


def normalize_channel_wise(tensor: torch.Tensor, mean: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
    """Normalizes given tensor channel-wise

    Parameters
    ----------
    tensor: torch.Tensor
        Tensor to be normalized
    mean: torch.tensor
        Mean to be subtracted
    std: torch.Tensor
        Std to be divided by

    Returns
    -------
    result: torch.Tensor

    """
    if len(tensor.size()) != 3:
        raise ValueError

    for channel in range(tensor.size(0)):
        tensor[channel, :, :] -= mean[channel]
        tensor[channel, :, :] /= std[channel]

    return tensor



import datetime
import pickle


class GlobalKVS(object):
    """
    Singleton dictionary with timestamps.
    Handy tool for data exchange in Machine Learning pipelines.

    """
    _instance = None
    _d = dict()
    _save_dir = None

    def __new__(cls, save_dir=None):
        if not cls._instance:
            cls._instance = super(GlobalKVS, cls).__new__(cls)
        if cls._instance._save_dir is None:
            cls._instance._save_dir = save_dir
        return cls._instance

    def set_save_dir(self, save_dir=None):
        self._save_dir = save_dir

    def update(self, tag, value, dtype=None, save_state=True):
        """
        Updates the internal state of the logger.

        Parameters
        ----------
        tag : str
            Tag, of the variable, which we log.
        value : object
            The value to be logged
        dtype :
            Container which is used to store the values under the tag
        save_state : bool
            Whether to pickle state
        Returns
        -------

        """
        if tag not in self._d:
            if dtype is None:
                self._d[tag] = (value, str(datetime.datetime.now()))
            else:
                self._d[tag] = dtype()
        else:
            if isinstance(self._d[tag], list):
                self._d[tag].append((value, str(datetime.datetime.now())))
            elif isinstance(self._d[tag], dict):
                self._d[tag].update((value, str(datetime.datetime.now())))
            else:
                self._d[tag] = (value, str(datetime.datetime.now()))
        if save_state:
            if self._save_dir is not None:
                self.save_pkl(self._save_dir)

    def __getitem__(self, tag):
        if not isinstance(self._d[tag], (list, dict)):
            return self._d[tag][0]
        else:
            return self._d[tag]

    def tag_ts(self, tag):
        return self._d[tag]

    def __contains__(self, key):
        return key in self._d

    def save_pkl(self, filename):
        with open(filename, 'wb') as f:
            pickle.dump(self._d, f)


from functools import partial
from torchvision import transforms as tvt

import solt.core as slc
import solt.data as sld
import solt.transforms as slt

from deeppipeline.kvs import GlobalKVS
from deeppipeline.common.transforms import numpy2tens, apply_by_index


def img_binary_mask2solt(imgmask):
    img, mask = imgmask
    if len(img.shape) == 2:
        img = img.reshape(img.shape[0], img.shape[1], 1)
    return sld.DataContainer((img, mask.squeeze()), 'IM')


def solt2img_binary_mask(dc: sld.DataContainer):
    if dc.data_format != 'IM':
        raise ValueError
    if not isinstance(dc, sld.DataContainer):
        raise TypeError

    return dc.data[0], dc.data[1]


def init_binary_segmentation_augs():
    kvs = GlobalKVS()
    ppl = tvt.Compose([
        img_binary_mask2solt,
        slc.Stream([
            slt.PadTransform(pad_to=(kvs['args'].pad_x, kvs['args'].pad_y)),
            slt.RandomFlip(axis=1, p=0.5),
            slt.CropTransform(crop_size=(kvs['args'].crop_x, kvs['args'].crop_y), crop_mode='r'),
            slt.ImageGammaCorrection(gamma_range=(kvs['args'].gamma_min, kvs['args'].gamma_max), p=0.5),
        ]),
        solt2img_binary_mask,
        partial(apply_by_index, transform=numpy2tens, idx=[0, 1]),
    ])

    kvs.update('train_trf', ppl)

    return ppl





from deeppipeline.kvs import GlobalKVS
from ._losses import CombinedLoss, BCEWithLogitsLoss2d, SoftJaccardLoss, FocalLoss


def init_binary_loss():
    kvs = GlobalKVS()
    if kvs['args'].n_classes == 2:
        if kvs['args'].loss == 'combined':
            return CombinedLoss([BCEWithLogitsLoss2d(),
                                 SoftJaccardLoss(use_log=kvs['args'].log_jaccard)],
                                weights=[1-kvs['args'].loss_weight,
                                kvs['args'].loss_weight])
        elif kvs['args'].loss == 'bce':
            return BCEWithLogitsLoss2d()
        elif kvs['args'].loss == 'jaccard':
            return SoftJaccardLoss(use_log=kvs['args'].log_jaccard)
        elif kvs['args'].loss == 'focal':
            return FocalLoss()
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError



import torch.nn as nn
import torch
import torch.nn.functional as F

from deeppipeline.common.modules import conv_block_3x3


class Encoder(nn.Module):
    """Encoder block. For encoder-decoder architecture.
    Conv3x3-Conv3x3-Maxpool 2x2. BatchNorm or other form of normalization is used optionally.

    Aleksei Tiulpin, Unversity of Oulu, 2017 (c).

    """

    def __init__(self, inp_channels, out_channels, depth=2, activation='relu', normalization='BN'):
        super().__init__()
        self.layers = nn.Sequential()

        for i in range(depth):
            tmp = []
            if i == 0:
                tmp.append(conv_block_3x3(inp_channels, out_channels, activation, normalization))
            else:
                tmp.append(conv_block_3x3(out_channels, out_channels, activation, normalization))
            self.layers.add_module('conv_3x3_{}'.format(i), nn.Sequential(*tmp))

    def forward(self, x):
        processed = self.layers(x)
        pooled = F.max_pool2d(processed, 2, 2)
        return processed, pooled


class Decoder(nn.Module):
    """Decoder block. For encoder-decoder architecture.
    Bilinear ups->Conv3x3-Conv3x3. BatchNorm or other form of normalization is used optionally.

    Aleksei Tiulpin, Unversity of Oulu, 2017 (c).

    """

    def __init__(self, inp_channels, out_channels, depth=2, mode='bilinear', activation='relu', normalization='BN'):
        super().__init__()
        self.layers = nn.Sequential()
        self.ups_mode = mode
        self.layers = nn.Sequential()

        for i in range(depth):
            tmp = []
            if i == 0:
                tmp.append(conv_block_3x3(inp_channels, out_channels, activation, normalization))
            else:
                tmp.append(conv_block_3x3(out_channels, out_channels, activation, normalization))
            self.layers.add_module('conv_3x3_{}'.format(i), nn.Sequential(*tmp))

    def forward(self, x_big, x):
        x_ups = F.interpolate(x, size=x_big.size()[-2:], mode=self.ups_mode, align_corners=True)
        y = torch.cat([x_ups, x_big], 1)
        y = self.layers(y)
        return y



import copy
import torch
import numpy as np


def apply_by_index(items, transform, idx=0):
    """Applies callable to certain objects in iterable using given indices.

    Parameters
    ----------
    items: tuple or list
    transform: callable
    idx: int or tuple or or list None
    Returns
    -------
    result: tuple

    """
    if idx is None:
        return items
    if not isinstance(items, (tuple, list)):
        raise TypeError
    if not isinstance(idx, (int, tuple, list)):
        raise TypeError

    if isinstance(idx, int):
        idx = [idx, ]

    idx = set(idx)
    res = []
    for i, item in enumerate(items):
        if i in idx:
            res.append(transform(item))
        else:
            res.append(copy.deepcopy(item))

    return res


def numpy2tens(x: np.ndarray, dtype='f') -> torch.Tensor:
    """Converts a numpy array into torch.Tensor

    Parameters
    ----------
    x: np.ndarray
        Array to be converted
    dtype: str
        Target data type of the tensor. Can be f - float and l - long

    Returns
    -------
    result: torch.Tensor

    """
    x = x.squeeze()
    x = torch.from_numpy(x)
    if x.dim() == 2:  # CxHxW format
        x = x.unsqueeze(0)

    if dtype == 'f':
        return x.float()
    elif dtype == 'l':
        return x.long()
    else:
        raise NotImplementedError



import cv2
import numpy as np


def read_gs_ocv(fname):
    return np.expand_dims(cv2.imread(fname, 0), -1)


def read_rgb_ocv(fname):
    return cv2.cvtColor(cv2.imread(fname, 0), cv2.COLOR_BGR2RGB)


def read_gs_binary_mask_ocv(fname):
    return np.expand_dims((cv2.imread(fname, 0) > 0).astype(np.float32), -1)


def read_3d_stack(slices):
    stack = None

    for i, fname in enumerate(slices):
        img = cv2.imread(fname, 0)

        if stack is None:
            stack = np.zeros((img.shape[0], len(slices), img.shape[1]), dtype=np.float16)
        stack[:, i, :] = img

    stack -= stack.min()
    stack /= stack.max()

    return stack



import torch.nn as nn
from collections import OrderedDict
from .modules import Encoder, Decoder
from deeppipeline.common.modules import conv_block_3x3


class UNet(nn.Module):
    """UNet architecture with 3x3 convolutions. Created dynamically based on the depth and width.

    Aleksei Tiulpin, Unversity of Oulu, 2017 (c).

    Parameters
    ----------
    bw : int
        Basic width of the network, which is doubled at each layer.
    depth : int
        Number of layers
    center_depth : int
        Depth of the central block in UNet.
    n_inputs : int
        Number of input channels.
    n_classes : int
        Number of input classes
    activation : str
        Activation function. Can be ReLU, SeLU, or ELU. The latter has parameter 1 by default.
    norm_encoder : str
        Normalization to be used in the encoder.
    norm_decoder : str
        Normalization to be used in the decoder.
    norm_center : str
        Normalization to be used in the center of the network.
    """

    def __init__(self, bw=24, depth=6, center_depth=2, n_inputs=1, n_classes=1,
                 activation='relu', norm_encoder='BN', norm_decoder='BN', norm_center='BN'):
        super().__init__()
        # Preparing the modules dict
        modules = OrderedDict()
        modules['down1'] = Encoder(n_inputs, bw, activation=activation, normalization=norm_encoder)
        # Automatically creating the Encoder based on the depth and width

        mul_out = None
        for level in range(2, depth + 1):
            mul_in = 2 ** (level - 2)
            mul_out = 2 ** (level - 1)
            layer = Encoder(bw * mul_in, bw * mul_out, activation=activation, normalization=norm_encoder)
            modules['down' + str(level)] = layer

        if mul_out is None:
            raise ValueError('The depth parameter is wrong. Cannot determine the output size of the encoder')
            # Creating the center
        modules['center'] = nn.Sequential(
            *[conv_block_3x3(bw * mul_out, bw * mul_out, activation, norm_center) for _ in range(center_depth)]
        )
        # Automatically creating the decoder
        for level in reversed(range(2, depth + 1)):
            mul_in = 2 ** (level - 1)
            layer = Decoder(2 * bw * mul_in, bw * mul_in // 2, activation=activation, normalization=norm_decoder)
            modules['up' + str(level)] = layer

        modules['up1'] = Decoder(bw + bw, bw, activation=activation, normalization=norm_decoder)

        modules['mixer'] = nn.Conv2d(bw, n_classes, kernel_size=1, padding=0, stride=1, bias=True)

        self.__dict__['_modules'] = modules

    def forward(self, x):
        encoded_results = {}
        out = x

        for name in self.__dict__['_modules']:
            if name.startswith('down'):
                layer = self.__dict__['_modules'][name]
                convolved, pooled = layer(out)
                encoded_results[name] = convolved
                out = pooled

        out = self.center(out)

        for name in self.__dict__['_modules']:
            if name.startswith('up'):
                layer = self.__dict__['_modules'][name]
                out = layer(encoded_results['down' + name[-1]], out)

        return self.mixer(out)



import numpy as np
import matplotlib.pyplot as plt


def cumulative_error_plot(errors, labels, title, colors=None, units='mm'):
    """
    Plots a cumulative curve. Useful for e.g. landmark detection.

    Parameters
    ----------
    errors : Array of errors
        Errors for the whole dataset for each landmark. Each sample in the dataset is represented by a row.
        The error for each landmark is stored in columns.
    labels : tuple or list of str or None
        Labels for each landmark
    title : str or None
        Title for the plot
    colors : list of str or None
        Colors for each landmark
    units : str
        Units to be displayed on X-axis

    Returns
    -------
    out : None
        Plots the cumulative curves.
    """
    plt.figure(figsize=(5, 5))
    for i in range(errors.shape[1]):
        sorted_data = np.sort(errors[:, i])
        if labels is not None:
            if colors is not None:
                plt.step(sorted_data, np.arange(sorted_data.size) / sorted_data.size, label=labels[i], colors=colors[i])
            else:
                plt.step(sorted_data, np.arange(sorted_data.size) / sorted_data.size, label=labels[i])
        else:
            if colors is not None:
                plt.step(sorted_data, np.arange(sorted_data.size) / sorted_data.size, color=colors[i])
            else:
                plt.step(sorted_data, np.arange(sorted_data.size) / sorted_data.size)

    plt.xlim(0, 4)
    plt.ylim(0, 1)
    plt.ylabel('Recall [%]')
    plt.xlabel(f'Distance from GT [f{units}]')
    plt.grid()
    if title is not None:
        plt.title(title)
    if labels is not None:
        plt.legend()
    plt.show()



import torch
from torch import nn
import torch.nn.functional as F
import math


class ElasticLoss(nn.Module):
    def __init__(self, w=0.5):
        super(ElasticLoss, self).__init__()
        self.weights = torch.FloatTensor([w, 1 - w])

    def forward(self, preds, gt):
        loss = 0

        if not isinstance(preds, tuple):
            preds = (preds, )

        for i in range(len(preds)):
            l2 = F.mse_loss(preds[i].squeeze(), gt.squeeze()).mul(self.weights[0])
            l1 = F.l1_loss(preds[i].squeeze(), gt.squeeze()).mul(self.weights[1])
            loss += l1 + l2

        return loss


class LNLoss(nn.Module):
    def __init__(self, space='l2'):
        super(LNLoss, self).__init__()
        if space == 'l2':
            self.loss = F.mse_loss
        elif space == 'l1':
            self.loss = F.l1_loss
        else:
            raise NotImplementedError

    def forward(self, preds, gt):
        loss = 0

        if not isinstance(preds, tuple):
            preds = (preds, )

        for i in range(len(preds)):
            loss += self.loss(preds[i].squeeze(), gt.squeeze())

        return loss


class WingLoss(nn.Module):
    """
    https://arxiv.org/pdf/1711.06753.pdf

    Refactored implementation from  https://github.com/BloodAxe/pytorch-toolbelt/.
    Supports intermediate supervision.

    """
    def __init__(self, width=5, curvature=0.5):
        super(WingLoss, self).__init__()
        self.w = width
        self.curvature = curvature
        self.c = self.w - self.w * math.log(1 + self.w / self.curvature)

    def __forward_single(self,preds, target):
        diff_abs = (target - preds).abs()
        loss = diff_abs.clone()

        idx_smaller = diff_abs < self.w
        idx_bigger = diff_abs >= self.w

        loss[idx_smaller] = self.w * torch.log(1 + diff_abs[idx_smaller] / self.curvature)
        loss[idx_bigger] = loss[idx_bigger] - self.c

        return loss.mean()

    def forward(self, preds, target):
        loss = 0

        if not isinstance(preds, tuple):
            preds = (preds, )

        for i in range(len(preds)):
            loss += self.__forward_single(preds[i], target)

        return loss




import torch
from torch.utils.data import DataLoader, sampler
import cv2
from tqdm import tqdm
import os
import glob
from termcolor import colored
from deeppipeline.segmentation.models import init_model
from deeppipeline.segmentation.training.dataset import SegmentationDataset
import numpy as np


def load_fold(args, fold_id):
    snapshot_name = glob.glob(os.path.join(args.snapshots_root, args.snapshot, f'fold_{fold_id}*.pth'))
    if len(snapshot_name) == 0:
        raise ValueError("Snapshot can't be found")
    snapshot_name = snapshot_name[0]

    net = init_model(ignore_data_parallel=True)
    snp = torch.load(snapshot_name)
    if isinstance(snp, dict):
        net_snp = snp['model']
    else:
        net_snp = snp
    net.load_state_dict(net_snp)

    if torch.cuda.device_count() > 1:
        net = torch.nn.DataParallel(net).to('cuda')
    net.eval()
    return net


def run_oof_binary(args, session_backup, read_img, read_mask, img_group_id_colname=None):
    metadata = session_backup[f'metadata'][0]
    if img_group_id_colname is not None:
        for group_name, _ in metadata.groupby(by=img_group_id_colname):
            os.makedirs(os.path.join(args.snapshots_root, args.snapshot, 'oof_inference', group_name), exist_ok=True)
    else:
        os.makedirs(os.path.join(args.snapshots_root, args.snapshot, 'oof_inference'), exist_ok=True)

    for fold_id, _, val_set in session_backup['cv_split'][0]:
        print(colored('====> ', 'green') + f'Loading fold [{fold_id}]')
        net = load_fold(args, fold_id)

        if args.tta:
            raise NotImplementedError('TTA is not yet supported')

        val_dataset = SegmentationDataset(split=val_set,
                                          trf=session_backup['val_trf'][0],
                                          read_img=read_img,
                                          read_mask=read_mask,
                                          img_group_id_colname=img_group_id_colname)

        val_loader = DataLoader(val_dataset, batch_size=args.bs,
                                num_workers=args.n_threads,
                                sampler=sampler.SequentialSampler(val_dataset))

        with torch.no_grad():
            for batch in tqdm(val_loader, total=len(val_loader), desc=f'Predicting fold {fold_id}:'):
                img = batch['img']
                if img_group_id_colname is not None:
                    group_ids = batch['group_id']
                else:
                    group_ids = None
                fnames = batch['fname']
                predicts = torch.sigmoid(net(img)).mul(255).to('cpu').numpy().astype(np.uint8)

                for idx, fname in enumerate(fnames):
                    pred_mask = predicts[idx].squeeze()
                    if img_group_id_colname is not None:
                        cv2.imwrite(os.path.join(args.snapshots_root,
                                                 args.snapshot,
                                                 'oof_inference', group_ids[idx], fname), pred_mask)
                    else:
                        cv2.imwrite(os.path.join(args.snapshots_root,
                                                 args.snapshot,
                                                 'oof_inference', fname), pred_mask)



import os
import pandas as pd
from sklearn.model_selection import KFold, GroupKFold, StratifiedKFold
from deeppipeline.kvs import GlobalKVS
import numpy as np


def init_folds(img_group_id_colname=None, img_class_colname=None):
    """
    Initialzies the cross-validation splits.

    Parameters
    ----------
    img_group_id_colname : str or None
        Column in `metadata` that is used to create cross-validation splits.
        If not None, then images that have the same group_id are never in train and validation.
    img_class_colname : str or None
        Column in `metadata` that is used to create cross-validation splits. If not none,
        splits are stratifed to ensure the same distribution of `img_class_colname` in train and validation.

    Returns
    -------

    """
    kvs = GlobalKVS()

    if img_group_id_colname is not None:
        gkf = GroupKFold(kvs['args'].n_folds)
        if img_class_colname is not None:
            class_col_name = getattr(kvs['metadata'], img_class_colname, None)
        else:
            class_col_name = None
        splitter = gkf.split(X=kvs['metadata'],
                             y=class_col_name,
                             groups=getattr(kvs['metadata'], img_group_id_colname))
    else:
        if img_class_colname is not None:
            skf = StratifiedKFold(kvs['args'].n_folds)
            splitter = skf.split(X=kvs['metadata'],
                                 y=getattr(kvs['metadata'], img_class_colname, None))
        else:
            kf = KFold(kvs['args'].n_folds)
            splitter = kf.split(X=kvs['metadata'])

    cv_split = []
    for fold_id, (train_ind, val_ind) in enumerate(splitter):

        if kvs['args'].fold != -1 and fold_id != kvs['args'].fold:
            continue

        np.random.shuffle(train_ind)
        train_ind = train_ind[::kvs['args'].skip_train]

        cv_split.append((fold_id,
                         kvs['metadata'].iloc[train_ind],
                         kvs['metadata'].iloc[val_ind]))

        kvs.update(f'losses_fold_[{fold_id}]', None, list)
        kvs.update(f'val_metrics_fold_[{fold_id}]', None, list)

    kvs.update('cv_split', cv_split)


def init_pd_meta():
    """
    Basic implementation of metadata loading. Loads the pandas data frame and stores
    it in global KVS under the `metadata` tag.

    Returns
    -------
    out : None
    """
    kvs = GlobalKVS()
    metadata = pd.read_csv(os.path.join(kvs['args'].workdir, kvs['args'].metadata))
    kvs.update('metadata', metadata)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


requirements = ('numpy', 'opencv-python',
                'torch>=1.0.0', 'solt>=0.1.5', 'torchvision', 'tqdm',
                'termcolor', 'tensorflow', 'tensorboardx', 'scikit-learn', 'pandas')

setup_requirements = ()

test_requirements = ('pytest',)

description = """A collection of building blocks for deep learning pipelines"""

setup(
    author="Aleksei Tiulpin",
    author_email='aleksei.tiulpin@oulu.fi',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux'
    ],
    description="Building blocks for Deep Learning",
    install_requires=requirements,
    license="MIT license",
    long_description=description,
    include_package_data=True,
    keywords='deeep learning, image segmentation, image-classification',
    name='deep-pipeline',
    packages=find_packages(include=['deeppipeline']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/lext/deep-pipeline',
    version='0.2.4',
    zip_safe=False,
)



from torch import nn


def conv_block_nxn(ks, inp, out, activation='relu', normalization='BN', bias=True):
    """3x3 ConvNet building block with different activations and normalizations support.

    Aleksei Tiulpin, Unversity of Oulu, 2017 (c).

    """
    if normalization == 'BN':
        norm_layer = nn.BatchNorm2d(out)
    elif normalization == 'IN':
        norm_layer = nn.InstanceNorm2d(out)
    elif normalization is None:
        norm_layer = nn.Sequential()
    else:
        raise NotImplementedError

    if activation == 'relu':
        return nn.Sequential(
            nn.Conv2d(inp, out, kernel_size=ks, padding=1, bias=bias),
            norm_layer,
            nn.ReLU(inplace=True)
        )

    elif activation == 'selu':
        return nn.Sequential(
            nn.Conv2d(inp, out, kernel_size=ks, padding=1, bias=bias),
            nn.SELU(inplace=True)
        )

    elif activation == 'elu':
        return nn.Sequential(
            nn.Conv2d(inp, out, kernel_size=ks, padding=1, bias=bias),
            nn.ELU(1, inplace=True)
        )
    elif activation is None:
        return nn.Sequential(
            nn.Conv2d(inp, out, kernel_size=ks, padding=1, bias=bias),
            norm_layer,
        )


def conv_block_3x3(inp, out, activation='relu', normalization='BN', bias=True):
    return conv_block_nxn(3, inp, out, activation, normalization, bias)


def conv_block_1x1(inp, out, activation='relu', normalization='BN', bias=True):
    return conv_block_nxn(3, inp, out, activation, normalization, bias)


class Identity(nn.Module):
    def forward(self, x):
        return x



from deeppipeline.common.modules import Identity, conv_block_3x3, conv_block_1x1
import torch
from torch import nn
import torch.nn.functional as F


class HGResidual(nn.Module):
    def __init__(self, n_inp, n_out):
        super().__init__()
        self.bottleneck = conv_block_1x1(n_inp, n_out // 2, 'relu')
        self.conv = conv_block_3x3(n_out // 2, n_out // 2, 'relu')
        self.out = conv_block_1x1(n_out // 2, n_out, None)

        if n_inp != n_out:
            self.skip = conv_block_1x1(n_inp, n_out, None)
        else:
            self.skip = Identity()

    def forward(self, x):
        o1 = self.bottleneck(x)
        o2 = self.conv(o1)
        o3 = self.out(o2)

        return o3 + self.skip(x)


class MultiScaleHGResidual(nn.Module):
    """
    https://arxiv.org/pdf/1808.04803.pdf

    """
    def __init__(self, n_inp, n_out):
        super().__init__()
        self.scale1 = conv_block_3x3(n_inp, n_out // 2, 'relu')
        self.scale2 = conv_block_3x3(n_out // 2, n_out // 4, 'relu')
        self.scale3 = conv_block_3x3(n_out // 4, n_out - n_out // 4 - n_out // 2, None)

        if n_inp != n_out:
            self.skip = conv_block_1x1(n_inp, n_out, None)
        else:
            self.skip = Identity()

    def forward(self, x):
        o1 = self.scale1(x)
        o2 = self.scale2(o1)
        o3 = self.scale3(o2)

        return torch.cat([o1, o2, o3], 1) + self.skip(x)


class SoftArgmax2D(nn.Module):
    def __init__(self, beta=1):
        super(SoftArgmax2D, self).__init__()
        self.beta = beta

    def forward(self, hm):
        hm = hm.mul(self.beta)
        bs, nc, h, w = hm.size()
        hm = hm.squeeze()

        softmax = F.softmax(hm.view(bs, nc, h*w), dim=2).view(bs, nc, h, w)

        weights = torch.ones(bs, nc, h, w).float().to(hm.device)
        w_x = torch.arange(w).float().div(w)
        w_x = w_x.to(hm.device).mul(weights)

        w_y = torch.arange(h).float().div(h)
        w_y = w_y.to(hm.device).mul(weights.transpose(2, 3)).transpose(2, 3)

        approx_x = softmax.mul(w_x).view(bs, nc, h*w).sum(2).unsqueeze(2)
        approx_y = softmax.mul(w_y).view(bs, nc, h*w).sum(2).unsqueeze(2)

        res_xy = torch.cat([approx_x, approx_y], 2)
        return res_xy


class Hourglass(nn.Module):
    def __init__(self, n, hg_width, n_inp, n_out, upmode='nearest', multiscale_block=False):
        super(Hourglass, self).__init__()

        self.multiscale_block = multiscale_block
        self.upmode = upmode

        self.lower1 = self.__make_block(n_inp, hg_width)
        self.lower2 = self.__make_block(hg_width, hg_width)
        self.lower3 = self.__make_block(hg_width, hg_width)

        if n > 1:
            self.lower4 = Hourglass(n - 1, hg_width, hg_width, n_out, upmode)
        else:
            self.lower4 = self.__make_block(hg_width, n_out)

        self.lower5 = self.__make_block(n_out, n_out)

        self.upper1 = self.__make_block(n_inp, hg_width)
        self.upper2 = self.__make_block(hg_width, hg_width)
        self.upper3 = self.__make_block(hg_width, n_out)

    def __make_block(self, inp, out):
        if self.multiscale_block:
            return MultiScaleHGResidual(inp, out)
        else:
            return HGResidual(inp, out)

    def forward(self, x):
        o_pooled = F.max_pool2d(x, 2)

        o1 = self.lower1(o_pooled)
        o2 = self.lower2(o1)
        o3 = self.lower3(o2)

        o4 = self.lower4(o3)

        o1_u = self.upper1(x)
        o2_u = self.upper2(o1_u)
        o3_u = self.upper3(o2_u)
        return o3_u + F.interpolate(self.lower5(o4), x.size()[-2:], mode=self.upmode, align_corners=True)




import torch
from torch.utils import data
import numpy as np

import os
from functools import partial
from torchvision import transforms as tvt

from deeppipeline.io import read_rgb_ocv, read_gs_binary_mask_ocv
from deeppipeline.kvs import GlobalKVS
from deeppipeline.common.normalization import init_mean_std, normalize_channel_wise
from deeppipeline.common.transforms import apply_by_index, numpy2tens


class SegmentationDataset(data.Dataset):
    def __init__(self, split, trf, read_img, read_mask, img_id_colname=None, img_group_id_colname=None):
        self.split = split
        self.transforms = trf
        self.read_img = read_img
        self.read_mask = read_mask
        self.img_id_colname = img_id_colname
        self.img_group_id_colname = img_group_id_colname

    def __getitem__(self, idx):
        entry = self.split.iloc[idx]
        img_fname = entry.img_fname
        mask_fname = entry.mask_fname

        img = self.read_img(img_fname)
        mask = self.read_mask(mask_fname)

        img, mask = self.transforms((img, mask))

        res = {'img': img, 'mask': mask, 'fname': img_fname.split('/')[-1]}
        if self.img_id_colname is not None:
            res['img_id'] = getattr(entry, self.img_id_colname)

        if self.img_group_id_colname is not None:
            res['group_id'] = getattr(entry, self.img_group_id_colname)

        return res

    def __len__(self):
        return self.split.shape[0]


def init_data_processing(img_reader=read_rgb_ocv, mask_reader=read_gs_binary_mask_ocv):
    kvs = GlobalKVS()

    dataset = SegmentationDataset(split=kvs['metadata'],
                                  trf=kvs['train_trf'],
                                  read_img=img_reader,
                                  read_mask=mask_reader)

    tmp = init_mean_std(snapshots_dir=os.path.join(kvs['args'].workdir, 'snapshots'),
                        dataset=dataset,
                        batch_size=kvs['args'].bs,
                        n_threads=kvs['args'].n_threads,
                        n_classes=kvs['args'].n_classes)

    if len(tmp) == 3:
        mean_vector, std_vector, class_weights = tmp
    elif len(tmp) == 2:
        mean_vector, std_vector = tmp
    else:

        raise ValueError('Incorrect format of mean/std/class-weights')

    norm_trf = partial(normalize_channel_wise, mean=mean_vector, std=std_vector)

    train_trf = tvt.Compose([
        kvs['train_trf'],
        partial(apply_by_index, transform=norm_trf, idx=0)
    ])

    val_trf = tvt.Compose([
        partial(apply_by_index, transform=numpy2tens, idx=[0, 1]),
        partial(apply_by_index, transform=norm_trf, idx=0)
    ])
    kvs.update('class_weights', class_weights)
    kvs.update('train_trf', train_trf)
    kvs.update('val_trf', val_trf)


def init_segmentation_loaders(x_train, x_val, img_reader=read_rgb_ocv, mask_reader=read_gs_binary_mask_ocv,
                              img_id_colname=None, img_group_id_colname=None):
    kvs = GlobalKVS()

    train_dataset = SegmentationDataset(split=x_train,
                                        trf=kvs['train_trf'],
                                        read_img=img_reader,
                                        read_mask=mask_reader,
                                        img_id_colname=img_id_colname,
                                        img_group_id_colname=img_group_id_colname)

    val_dataset = SegmentationDataset(split=x_val,
                                      trf=kvs['val_trf'],
                                      read_img=img_reader,
                                      read_mask=mask_reader,
                                      img_id_colname=img_id_colname,
                                      img_group_id_colname=img_group_id_colname)

    train_loader = data.DataLoader(train_dataset, batch_size=kvs['args'].bs,
                                   num_workers=kvs['args'].n_threads, shuffle=True,
                                   drop_last=True,
                                   worker_init_fn=lambda wid: np.random.seed(np.uint32(torch.initial_seed() + wid)))

    val_loader = data.DataLoader(val_dataset, batch_size=kvs['args'].val_bs,
                                 num_workers=kvs['args'].n_threads)

    return train_loader, val_loader




import torch
from tqdm import tqdm
import gc
import numpy as np
import os
from tensorboardX import SummaryWriter
from termcolor import colored

from deeppipeline.common.core import save_checkpoint, init_optimizer, init_session
from deeppipeline.common.dataset import init_folds
from deeppipeline.kvs import GlobalKVS

from deeppipeline.segmentation.models import init_model
from deeppipeline.segmentation.losses import init_binary_loss
from deeppipeline.segmentation.training.dataset import init_segmentation_loaders
from deeppipeline.segmentation.training.dataset import init_data_processing

from deeppipeline.segmentation.evaluation import metrics
from deeppipeline.segmentation.evaluation.metrics import calculate_dice, calculate_iou
from deeppipeline.transforms.segmentation import init_binary_segmentation_augs

def pass_epoch(net, loader, optimizer, criterion):
    kvs = GlobalKVS()
    net.train(optimizer is not None)

    fold_id = kvs['cur_fold']
    epoch = kvs['cur_epoch']
    max_ep = kvs['args'].n_epochs
    n_classes = kvs['args'].n_classes

    running_loss = 0.0
    n_batches = len(loader)
    confusion_matrix = np.zeros((n_classes, n_classes), dtype=np.uint64)
    device = next(net.parameters()).device
    pbar = tqdm(total=n_batches, ncols=200)
    with torch.set_grad_enabled(optimizer is not None):
        for i, entry in enumerate(loader):
            if optimizer is not None:
                optimizer.zero_grad()

            inputs = entry['img'].to(device)
            mask = entry['mask'].to(device)
            outputs = net(inputs)
            loss = criterion(outputs, mask)

            if optimizer is not None:
                loss.backward()
                optimizer.step()
                running_loss += loss.item()
                pbar.set_description(f"Fold [{fold_id}] [{epoch} | {max_ep}] | "
                                     f"Running loss {running_loss / (i + 1):.5f} / {loss.item():.5f}")
            else:
                running_loss += loss.item()
                pbar.set_description(desc=f"Fold [{fold_id}] [{epoch} | {max_ep}] | Validation progress")

                if n_classes == 2:
                    preds = outputs.gt(kvs['args'].binary_threshold)
                elif n_classes > 2:
                    preds = outputs.argmax(axis=1)
                else:
                    raise ValueError

                preds = preds.float().to('cpu').numpy()
                mask = mask.float().to('cpu').numpy()
                confusion_matrix += metrics.calculate_confusion_matrix_from_arrays(preds, mask, n_classes)

            pbar.update()
            gc.collect()
        gc.collect()
        pbar.close()

    return running_loss / n_batches, confusion_matrix


def log_metrics(writer, train_loss, val_loss, conf_matrix):
    kvs = GlobalKVS()

    dices = {'dice_{}'.format(cls): dice for cls, dice in enumerate(calculate_dice(conf_matrix))}
    ious = {'iou_{}'.format(cls): iou for cls, iou in enumerate(calculate_iou(conf_matrix))}
    print(colored('==> ', 'green') + 'Metrics:')
    print(colored('====> ', 'green') + 'Train loss:', train_loss)
    print(colored('====> ', 'green') + 'Val loss:', val_loss)
    print(colored('====> ', 'green') + f'Val Dice: {dices}')
    print(colored('====> ', 'green') + f'Val IoU: {ious}')
    dices_tb = {}
    for cls in range(1, len(dices)):
        dices_tb[f"Dice [{cls}]"] = dices[f"dice_{cls}"]

    ious_tb = {}
    for cls in range(1, len(ious)):
        ious_tb[f"IoU [{cls}]"] = ious[f"iou_{cls}"]

    to_log = {'train_loss': train_loss, 'val_loss': val_loss}
    # Tensorboard logging
    writer.add_scalars(f"Losses_{kvs['args'].model}", to_log, kvs['cur_epoch'])
    writer.add_scalars('Metrics/Dice', dices_tb, kvs['cur_epoch'])
    writer.add_scalars('Metrics/IoU', ious_tb, kvs['cur_epoch'])
    # KVS logging
    to_log.update({'epoch': kvs['cur_epoch']})
    val_metrics = {'epoch': kvs['cur_epoch']}
    val_metrics.update(to_log)
    val_metrics.update(dices)
    val_metrics.update({'conf_matrix': conf_matrix})

    kvs.update(f'losses_fold_[{kvs["cur_fold"]}]', to_log)
    kvs.update(f'val_metrics_fold_[{kvs["cur_fold"]}]', val_metrics)


def train_fold(net, train_loader, optimizer, criterion, val_loader, scheduler):
    kvs = GlobalKVS()
    fold_id = kvs['cur_fold']
    writer = SummaryWriter(os.path.join(kvs['args'].workdir, 'snapshots', kvs['snapshot_name'],
                                        'logs', 'fold_{}'.format(fold_id), kvs['snapshot_name']))

    for epoch in range(kvs['args'].n_epochs):
        print(colored('==> ', 'green') + f'Training epoch [{epoch}] with LR {scheduler.get_lr()}')
        kvs.update('cur_epoch', epoch)
        train_loss, _ = pass_epoch(net, train_loader, optimizer, criterion)
        val_loss, conf_matrix = pass_epoch(net, val_loader, None, criterion)
        log_metrics(writer, train_loss, val_loss, conf_matrix)
        save_checkpoint(net, optimizer, 'val_loss', 'lt')
        scheduler.step()


def train_n_folds(init_args, init_metadata, init_scheduler, img_reader,
                  mask_reader, init_augs=None, img_group_id_colname=None, img_class_colname=None, img_id_colname=None):

    args = init_args()
    kvs = init_session(args)[-1]
    init_metadata()
    assert 'metadata' in kvs
    if init_augs is None:
        if kvs['args'].n_classes == 2:
            init_augs = init_binary_segmentation_augs()
        else:
            raise NotImplementedError('Augmentations are not defined')
    else:
        init_augs()
    assert 'train_trf' in kvs
    init_data_processing(img_reader=img_reader, mask_reader=mask_reader)
    init_folds(img_group_id_colname=img_group_id_colname, img_class_colname=img_class_colname)

    for fold_id, x_train, x_val in kvs['cv_split']:
        kvs.update('cur_fold', fold_id)
        kvs.update('prev_model', None)

        net = init_model()
        optimizer = init_optimizer(net)
        if kvs['args'].n_classes == 2:
            criterion = init_binary_loss()
        else:
            raise NotImplementedError('Loss is not defined')
        scheduler = init_scheduler(optimizer)
        train_loader, val_loader = init_segmentation_loaders(x_train=x_train, x_val=x_val,
                                                             img_reader=img_reader,
                                                             mask_reader=mask_reader,
                                                             img_id_colname=img_id_colname,
                                                             img_group_id_colname=img_group_id_colname)

        train_fold(net=net, train_loader=train_loader,
                   optimizer=optimizer, criterion=criterion,
                   val_loader=val_loader, scheduler=scheduler)

